import time
import multiprocessing as mp
from datetime import UTC, datetime

import astropy
import numpy as np
from dateutil.parser import parse as parse_datetime_str
from ndcube import NDCube
from numpy.polynomial import polynomial
from prefect import get_run_logger
from quadprog import solve_qp
from scipy.interpolate import griddata

from punchbowl.data import NormalizedMetadata, load_ndcube_from_fits
from punchbowl.data.wcs import load_trefoil_wcs
from punchbowl.exceptions import InvalidDataError
from punchbowl.prefect import punch_flow, punch_task
from punchbowl.util import nan_percentile


def solve_qp_cube(input_vals: np.ndarray, cube: np.ndarray,
                  n_nonnan_required: int=7) -> (np.ndarray, np.ndarray):
    """
    Fast solver for the quadratic programming problem.

    Parameters
    ----------
    input_vals : np.ndarray
        array of times
    cube : np.ndarray
        array of data
    n_nonnan_required : int
        The number of non-nan values that must be present in each pixel's time series.
        Any pixels with fewer will not be fit, with zeros returned instead.

    Returns
    -------
    np.ndarray
        Array of coefficients for solving polynomial

    """
    logger = get_run_logger()
    c = np.transpose(input_vals)
    cube_is_good = np.isfinite(cube)
    num_inputs = np.sum(cube_is_good, axis=0)

    solution = np.zeros((input_vals.shape[1], cube.shape[1], cube.shape[2]))
    for i in range(cube.shape[1]):
        start = time.time()
        for j in range(cube.shape[2]):
            is_good = cube_is_good[:, i, j]
            time_series = cube[:, i, j][is_good]
            if time_series.size < n_nonnan_required:
                solution[:, i, j] = 0
            else:
                c_iter = c[:, is_good]
                g_iter = np.matmul(c_iter, c_iter.T)
                a = np.matmul(c_iter, time_series)
                try:
                    solution[:, i, j] = solve_qp(g_iter, a, c_iter, time_series)[0]
                except ValueError:
                    solution[:, i, j] = 0
        end = time.time()
        logger.info(f"{i} took {end - start} seconds")
    return np.asarray(solution), num_inputs


def model_fcorona_for_cube(xt: np.ndarray,
                           reference_xt: float,
                           cube: np.ndarray,
                           min_brightness: float = 1E-18,
                           clip_factor: float | None = 1,
                           return_full_curves: bool = False,
                           num_workers: int | None = 8,
                           ) -> tuple[np.ndarray, np.ndarray] | tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Model the F corona given a list of times and a corresponding data cube.

    Parameters
    ----------
    xt : np.ndarray
        time array
    reference_xt: float
        timestamp to evaluate the model for
    cube : np.ndarray
        observation array
    min_brightness: float
        pixels dimmer than this value are set to nan and considered empty
    clip_factor : float | None
        If None, no smoothing is applied.
        Otherwise, the difference between the 25th and 75th percentile is computed and values that vary from the median
        by more than `clip_factor` times the difference data are rejected.
    return_full_curves: bool
        If True, this function returns the full curve fitted to the time series at each pixel
        and the smoothed data cube. If False (default), only the curve's value at the central
        frame is returned, producing a model at one instant in time.
    num_workers: int | None
        Work is parallelized over this many worker processes. If None, this matches the number of cores.

    Returns
    -------
    np.ndarray
        The F-corona model at the central point in time. If return_full_curves is True, this is
        instead the F-corona model at all points in time covered by the data cube
    np.ndarray
        The number of data points used in solving the F-corona model for each pixel of the output
    np.ndarray
        The smoothed data cube. Returned only if return_full_curves is True.

    """
    if num_workers < 2:
        # Skip forking, reassembling, etc.
        return _model_fcorona_for_cube_inner(xt, reference_xt, cube, min_brightness, clip_factor, return_full_curves)

    def args() -> tuple:
        # Generate a set of args for one task
        for i in range(cube.shape[1]):
            yield xt, reference_xt, cube[:, i:i + 1], min_brightness, clip_factor, return_full_curves

    with mp.Pool(processes=num_workers) as pool:
        chunks = pool.starmap(_model_fcorona_for_cube_inner, args(), chunksize=10)

    # Combine the outputs of each task into final output arrays
    if return_full_curves:
        curves, counts, cubes = zip(*chunks, strict=False)
        curves = np.concatenate(curves, axis=-2)
        counts = np.concatenate(counts, axis=-2)
        cubes = np.concatenate(cubes, axis=-2)
        return curves, counts, cubes
    model, counts = zip(*chunks, strict=False)
    model = np.concatenate(model, axis=-2)
    counts = np.concatenate(counts, axis=-2)
    return model, counts


def _model_fcorona_for_cube_inner(xt: np.ndarray,
                                  reference_xt: float,
                                  cube: np.ndarray,
                                  min_brightness: float = 1E-18,
                                  clip_factor: float | None = 1,
                                  return_full_curves: bool=False,
                                  ) -> tuple[np.ndarray, np.ndarray] | tuple[np.ndarray, np.ndarray, np.ndarray]:
    logger = get_run_logger()
    cube[cube < min_brightness] = np.nan
    if clip_factor is not None:
        low, center, high = nan_percentile(cube, [25, 50, 75])
        width = high - low
        a, b, c = np.where(cube[:, ...] > (center + (clip_factor * width)))
        cube[a, b, c] = np.nan

        a, b, c = np.where(cube[:, ...] < (center - (clip_factor * width)))
        cube[a, b, c] = np.nan

    xt = np.array(xt)
    reference_xt -= xt[0]
    xt -= xt[0]

    input_array = np.c_[np.power(xt, 3), np.square(xt), xt, np.ones(len(xt))]
    coefficients, counts = solve_qp_cube(input_array, -cube)
    coefficients *= -1
    if return_full_curves:
        return polynomial.polyval(xt, coefficients[::-1, :, :]).transpose((2, 0, 1)), counts, cube

    start = time.time()
    out = polynomial.polyval(reference_xt, coefficients[::-1, :, :]), counts
    end = time.time()
    logger.info(f"computed polyval in {end - start} seconds")
    return out


def fill_nans_with_interpolation(image: np.ndarray) -> np.ndarray:
    """Fill NaN values in an image using interpolation."""
    mask = np.isnan(image)
    x, y = np.where(~mask)
    known_values = image[~mask]

    grid_x, grid_y = np.mgrid[0:image.shape[0], 0:image.shape[1]]
    return griddata((x, y), known_values, (grid_x, grid_y), method="cubic")


def _load_one_file(filename: str) -> NDCube:
    """
    Support for loading files in parallel via multiprocessing.

    Parameters
    ----------
    filename : str
        The file to load

    Returns
    -------
    data_object : `NDCube`
        The loaded data

    """
    cube = load_ndcube_from_fits(filename)
    data = np.where(np.isnan(cube.uncertainty.array), np.nan, cube.data)
    return data, cube.meta


@punch_flow(log_prints=True)
def construct_f_corona_model(filenames: list[str],
                             clip_factor: float = 3.0,
                             reference_time: str | None = None,
                             num_workers: int = 8,
                             fill_nans: bool = True,
                             polarized: bool = False) -> list[NDCube]:
    """Construct a full F corona model."""
    logger = get_run_logger()

    if reference_time is None:
        reference_time = datetime.now(UTC)
    elif isinstance(reference_time, str):
        reference_time = parse_datetime_str(reference_time)

    trefoil_wcs, trefoil_shape = load_trefoil_wcs()

    logger.info("construct_f_corona_background started")

    if len(filenames) == 0:
        msg = "Require at least one input file"
        raise ValueError(msg)

    filenames.sort()

    data_shape = (3, *trefoil_shape) if polarized else trefoil_shape

    number_of_data_frames = len(filenames)
    data_cube = np.empty((number_of_data_frames, *data_shape), dtype=float)

    meta_list = []
    obs_times = []

    logger.info("beginning data loading")
    dates = []
    with mp.Pool(processes=num_workers) as pool:
        for i, (data, meta) in enumerate(pool.imap(_load_one_file, filenames)):
            dates.append(meta.datetime)
            data_cube[i] = data
            obs_times.append(meta.datetime.timestamp())
            meta_list.append(meta)
    logger.info("ending data loading")
    output_datebeg = min(dates).isoformat()
    output_dateend = max(dates).isoformat()

    reference_xt = reference_time.timestamp()
    if polarized:
        m_model_fcorona, _ = model_fcorona_for_cube(obs_times, reference_xt,
                                                    data_cube[:, 0, :, :],
                                                    num_workers=num_workers,
                                                    clip_factor=clip_factor)
        m_model_fcorona[m_model_fcorona==0] = np.nan
        if fill_nans:
            m_model_fcorona = fill_nans_with_interpolation(m_model_fcorona)

        z_model_fcorona, _ = model_fcorona_for_cube(obs_times,
                                                    reference_xt,
                                                    data_cube[:, 1, :, :],
                                                    num_workers=num_workers,
                                                    clip_factor=clip_factor)
        z_model_fcorona[z_model_fcorona==0] = np.nan
        if fill_nans:
            z_model_fcorona = fill_nans_with_interpolation(z_model_fcorona)

        p_model_fcorona, _ = model_fcorona_for_cube(obs_times,
                                                    reference_xt,
                                                    data_cube[:, 2, :, :],
                                                    num_workers=num_workers,
                                                    clip_factor=clip_factor)
        p_model_fcorona[p_model_fcorona==0] = np.nan
        if fill_nans:
            p_model_fcorona = fill_nans_with_interpolation(p_model_fcorona)

        output_data = np.stack([m_model_fcorona,
                                z_model_fcorona,
                                p_model_fcorona], axis=0)
        meta = NormalizedMetadata.load_template("PFM", "3")
        trefoil_wcs = astropy.wcs.utils.add_stokes_axis_to_wcs(trefoil_wcs, 2)
    else:
        model_fcorona, _ = model_fcorona_for_cube(obs_times, reference_xt,
                                                  data_cube,
                                                  num_workers=num_workers,
                                                  clip_factor=clip_factor)
        model_fcorona[model_fcorona==0] = np.nan
        if fill_nans:
            model_fcorona = fill_nans_with_interpolation(model_fcorona)

        output_data = model_fcorona
        meta = NormalizedMetadata.load_template("CFM", "3")

    meta["DATE"] = datetime.now(UTC).isoformat()
    meta["DATE-AVG"] = reference_time.isoformat()
    meta["DATE-OBS"] = reference_time.isoformat()

    meta["DATE-BEG"] = output_datebeg
    meta["DATE-END"] = output_dateend

    output_cube = NDCube(data=output_data,
                         meta=meta,
                         wcs=trefoil_wcs)


    return [output_cube]

def subtract_f_corona_background(data_object: NDCube,
                                 before_f_background_model: NDCube,
                                 after_f_background_model: NDCube ) -> NDCube:
    """Subtract f corona background."""
    # check dimensions match
    if data_object.data.shape != before_f_background_model.data.shape:
        msg = (
            "f_background_subtraction expects the data_object and"
            "f_background arrays to have the same dimensions."
            f"data_array dims: {data_object.data.shape} "
            f"and before_f_background_model dims: {before_f_background_model.data.shape}"
        )
        raise InvalidDataError(
            msg,
        )

    if data_object.data.shape != after_f_background_model.data.shape:
        msg = (
            "f_background_subtraction expects the data_object and"
            "f_background arrays to have the same dimensions."
            f"data_array dims: {data_object.data.shape} "
            f"and after_f_background_model dims: {after_f_background_model.data.shape}"
        )
        raise InvalidDataError(
            msg,
        )

    before_date = before_f_background_model.meta.datetime.timestamp()
    after_date = after_f_background_model.meta.datetime.timestamp()
    observation_date = data_object.meta.datetime.timestamp()

    if before_date > observation_date:
        msg = "Before F corona model was after the observation date"
        raise InvalidDataError(msg)

    if after_date < observation_date:
        msg = "After F corona model was before the observation date"
        raise InvalidDataError(msg)

    if before_date == observation_date:
        interpolated_model = before_f_background_model
    elif after_date == observation_date:
        interpolated_model = after_f_background_model
    else:
        interpolated_model = ((after_f_background_model.data - before_f_background_model.data)
                              * (observation_date - before_date) / (after_date - before_date)
                              + before_f_background_model.data)

    interpolated_model[np.isinf(data_object.uncertainty.array)] = 0

    original_mask = data_object.data[...] == 0
    data_object.data[...] = data_object.data[...] - interpolated_model
    data_object.data[original_mask] = 0
    data_object.uncertainty.array[:, :] -= interpolated_model
    data_object.uncertainty.array[original_mask] = 0
    return data_object

@punch_task
def subtract_f_corona_background_task(observation: NDCube,
                                      before_f_background_model_path: str,
                                      after_f_background_model_path: str) -> NDCube:
    """
    Subtracts a background f corona model from an observation.

    This algorithm linearly interpolates between the before and after models.

    Parameters
    ----------
    observation : NDCube
        an observation to subtract an f corona model from

    before_f_background_model_path : str
        path to a NDCube f corona background map before the observation

    after_f_background_model_path : str
        path to a NDCube f corona background map after the observation

    Returns
    -------
    NDCube
        A background subtracted data frame

    """
    logger = get_run_logger()
    logger.info("subtract_f_corona_background started")


    before_f_corona_model = load_ndcube_from_fits(before_f_background_model_path)
    after_f_corona_model = load_ndcube_from_fits(after_f_background_model_path)

    output = subtract_f_corona_background(observation, before_f_corona_model, after_f_corona_model)
    output.meta.history.add_now("LEVEL3-subtract_f_corona_background", "subtracted f corona background")

    logger.info("subtract_f_corona_background finished")

    return output


def create_empty_f_background_model(data_object: NDCube) -> np.ndarray:
    """Create an empty background model."""
    return np.zeros_like(data_object.data)



from math import floor
from datetime import UTC, datetime

import astropy.units as u
import numpy as np
import remove_starfield
from astropy.wcs import WCS
from dateutil.parser import parse as parse_datetime_str
from ndcube import NDCollection, NDCube
from prefect import get_run_logger
from remove_starfield import ImageHolder, ImageProcessor, Starfield
from remove_starfield.reducers import PercentileReducer
from solpolpy import resolve

from punchbowl.data import NormalizedMetadata, load_ndcube_from_fits
from punchbowl.data.wcs import calculate_helio_wcs_from_celestial, get_p_angle
from punchbowl.prefect import punch_flow, punch_task


def to_celestial(input_data: NDCube) -> NDCube:
    """
    Convert polarization from mzpsolar to Celestial frame.

    All images need their polarization converted to Celestial frame
    to generate the background starfield model.
    """
    # Create a data collection for M, Z, P components
    mzp_angles = [-60, 0, 60]*u.degree

    # Compute new angles for celestial frame
    cel_north_offset = get_p_angle(time=input_data[0].meta["DATE-OBS"].value)
    new_angles = mzp_angles - cel_north_offset

    collection_contents = [
        (label,
         NDCube(data=input_data[i].data,
                wcs=input_data.wcs.dropaxis(2),
                meta={"POLAR": angle}))
        for label, i, angle in zip(["M", "Z", "P"], [0, 1, 2], mzp_angles, strict=False)
    ]
    data_collection = NDCollection(collection_contents, aligned_axes="all")

    # Resolve data to celestial frame
    celestial_data_collection = resolve(data_collection, "npol", out_angles=new_angles, imax_effect=False)

    valid_keys = [key for key in celestial_data_collection if key != "alpha"]
    new_data = [celestial_data_collection[key].data for key in valid_keys]
    new_wcs = input_data.wcs.copy()

    output_meta = NormalizedMetadata.load_template("PTM", "3")
    output_meta["DATE-OBS"] = input_data.meta["DATE-OBS"].value

    output = NDCube(data=new_data, wcs=new_wcs, meta=output_meta)
    output.meta.history.add_now("LEVEL3-convert2celestial", "Convert mzpsolar to Celestial")

    return output


def from_celestial(input_data: NDCube) -> NDCube:
    """
    Convert polarization from Celestial frame to mzpsolar.

    All images need their polarization converted back to Solar frame
    after removing the stellar polarization.
    """
    # Create a data collection for M, Z, P components
    mzp_angles = [-60, 0, 60]*u.degree
    # Compute new angles for celestial frame
    cel_north_offset = get_p_angle(time=input_data[0].meta["DATE-OBS"].value)
    new_angles = mzp_angles - cel_north_offset
    collection_contents = [
        (f"{angle.value} deg",
         NDCube(data=input_data[i].data,
                wcs=input_data.wcs.dropaxis(2),
                meta={"POLAR": angle}))
        for i, angle in enumerate(new_angles)
    ]
    data_collection = NDCollection(collection_contents, aligned_axes="all")

    # Resolve data to mzpsolar frame
    solar_data_collection = resolve(data_collection, "mzpsolar", imax_effect=False)

    valid_keys = [key for key in solar_data_collection if key != "alpha"]
    new_data = [solar_data_collection[key].data for key in valid_keys]
    new_wcs = input_data.wcs.copy()

    output_meta = NormalizedMetadata.load_template("PTM", "2")
    output_meta["DATE-OBS"] = input_data.meta["DATE-OBS"].value

    output = NDCube(data=new_data, wcs=new_wcs, meta=output_meta, uncertainty=input_data.uncertainty)
    output.meta.history.add_now("LEVEL3-convert2mzpsolar", "Convert Celestial to mzpsolar")

    return output


class PUNCHImageProcessor(ImageProcessor):
    """Special loader for PUNCH data."""

    def __init__(self, layer: int, apply_mask: bool = True, key: str = " ") -> None:
        """Create PUNCHImageProcessor."""
        self.layer = layer
        self.apply_mask = apply_mask
        self.key = key

    def load_image(self, filename: str) -> ImageHolder:
        """Load an image."""
        cube = load_ndcube_from_fits(filename, key=self.key)
        cube = to_celestial(cube)
        data = cube.data[self.layer]
        if self.apply_mask:
            data[np.isclose(cube.uncertainty.array[self.layer], 0, atol=1E-30)] = np.nan
        return ImageHolder(data, cube.wcs.celestial, cube.meta)


@punch_flow(log_prints=True, timeout_seconds=21_600)
def generate_starfield_background(
        filenames: list[str],
        map_scale: float = 0.01,
        target_mem_usage: float = 1000,
        n_procs: int | None = None,
        reference_time: datetime | None = None) -> [NDCube, NDCube]:
    """Create a background starfield_bg map from a series of PUNCH images over a long period of time."""
    logger = get_run_logger()

    if reference_time is None:
        reference_time = datetime.now(UTC)
    elif isinstance(reference_time, str):
        reference_time = parse_datetime_str(reference_time)

    logger.info("construct_starfield_background started")

    # create an empty array to fill with data
    #   open the first file in the list to ge the shape of the file
    if len(filenames) == 0:
        msg = "filenames cannot be empty"
        raise ValueError(msg)

    shape = [floor(132 / map_scale), floor(360 / map_scale)]
    starfield_wcs = WCS(naxis=2)
    # n.b. it seems the RA wrap point is chosen so there's 180 degrees
    # included on either side of crpix
    crpix = [shape[1] / 2 + .5, shape[0] / 2 + .5]
    starfield_wcs.wcs.crpix = crpix
    starfield_wcs.wcs.crval = 270, -23.5
    starfield_wcs.wcs.cdelt = map_scale, map_scale
    starfield_wcs.wcs.ctype = "RA---CAR", "DEC--CAR"
    starfield_wcs.wcs.cunit = "deg", "deg"
    starfield_wcs.array_shape = shape

    logger.info("Starting m starfield")
    starfield_m = remove_starfield.build_starfield_estimate(
        filenames,
        attribution=False,
        frame_count=False,
        reducer=PercentileReducer(10),
        starfield_wcs=starfield_wcs,
        n_procs=n_procs,
        processor=PUNCHImageProcessor(0, apply_mask=True, key="A"),
        target_mem_usage=target_mem_usage)
    logger.info("Ending m starfield")

    logger.info("Starting z starfield")
    starfield_z = remove_starfield.build_starfield_estimate(
        filenames,
        attribution=False,
        frame_count=False,
        reducer=PercentileReducer(10),
        starfield_wcs=starfield_wcs,
        n_procs=n_procs,
        processor=PUNCHImageProcessor(1, apply_mask=True, key="A"),
        target_mem_usage=target_mem_usage)
    logger.info("Ending z starfield")


    logger.info("Starting p starfield")
    starfield_p = remove_starfield.build_starfield_estimate(
        filenames,
        attribution=False,
        frame_count=False,
        reducer=PercentileReducer(10),
        starfield_wcs=starfield_wcs,
        n_procs=n_procs,
        processor=PUNCHImageProcessor(2, apply_mask=True, key="A"),
        target_mem_usage=target_mem_usage)
    logger.info("Ending p starfield")

    # create an output PUNCHdata object
    logger.info("Preparing to create outputs")

    meta = NormalizedMetadata.load_template("PSM", "3")
    meta["DATE-OBS"] = reference_time.isoformat()
    meta["DATE-BEG"] = reference_time.isoformat()
    meta["DATE-END"] = reference_time.isoformat()
    meta["DATE-AVG"] = reference_time.isoformat()
    meta["DATE"] = datetime.now(UTC).isoformat()

    out_wcs, _ = calculate_helio_wcs_from_celestial(starfield_m.wcs, meta.astropy_time, starfield_m.starfield.shape)
    output = NDCube(np.stack([starfield_m.starfield, starfield_z.starfield, starfield_p.starfield], axis=0),
                    wcs=out_wcs, meta=meta)
    output.meta.history.add_now("LEVEL3-starfield_background", "constructed starfield_bg model")

    logger.info("construct_starfield_background finished")
    return [output]


@punch_task
def subtract_starfield_background_task(data_object: NDCube,
                                       starfield_background_path: str | None) -> NDCube:
    """
    Subtracts a background starfield from an input data frame.

    checks the dimensions of input data frame and background starfield match and
    subtracts the background starfield from the data frame of interest.

    Parameters
    ----------
    data_object : NDCube
        A NDCube data frame to be background subtracted

    starfield_background_path : str
        path to a NDCube background starfield map

    Returns
    -------
    NDCube
        A background starfield subtracted data frame

    """
    logger = get_run_logger()
    logger.info("subtract_starfield_background started")

    if starfield_background_path is None:
        output = data_object
        output.meta.history.add_now("LEVEL3-subtract_starfield_background",
                                           "starfield subtraction skipped since path is empty")
    else:
        star_datacube = load_ndcube_from_fits(starfield_background_path, key="A")
        # data_wcs = calculate_celestial_wcs_from_helio(data_object.wcs.celestial,
        #                                               data_object.meta.astropy_time,
        #                                               data_object.data.shape[-2:])

        starfield_model = Starfield(np.stack((star_datacube.data, star_datacube.uncertainty.array)), star_datacube.wcs)

        subtracted = starfield_model.subtract_from_image(
            NDCube(data=np.stack((data_object.data, data_object.uncertainty.array)),
                   wcs=data_object.wcs.celestial,
                   meta=data_object.meta),
            processor=PUNCHImageProcessor(0, key="A"))

        data_object.data[...] = subtracted.subtracted[0]
        data_object.uncertainty.array[...] -= subtracted.subtracted[1]
        data_object.meta.history.add_now("LEVEL3-subtract_starfield_background", "subtracted starfield background")
        output = from_celestial(data_object)
    logger.info("subtract_starfield_background finished")

    return output


def create_empty_starfield_background(data_object: NDCube) -> np.ndarray:
    """Create an empty starfield background map."""
    return np.zeros_like(data_object.data)



import numpy as np
from ndcube import NDCube
from scipy.signal import convolve2d, medfilt2d

from punchbowl.level1.deficient_pixel import cell_neighbors
from punchbowl.prefect import punch_task


def radial_array(shape: tuple[int], center: tuple[int] | None = None) -> np.ndarray:
    """Create radial array."""
    if len(shape) != 2:
        msg = f"Shape must be 2D, received {shape} with {len(shape)} dimensions"
        raise ValueError(msg)

    x = np.arange(shape[0])
    y = np.arange(shape[1])
    X, Y = np.meshgrid(x, y)  # noqa: N806

    center = [s // 2 for s in shape] if center is None else center
    return np.floor(np.sqrt(np.square(X - center[0]) + np.square(Y - center[1])))



def spikejones(
    image: np.ndarray, unsharp_size: int = 3, method: str = "convolve", alpha: float = 1, dilation: int = 0,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Remove cosmic ray spikes from an image using spikejones algorithm.

    This code is based on https://github.com/drzowie/solarpdl-tools/blob/master/image/spikejones.pdl

    Parameters
    ----------
    image : np.ndarray
        an array representing an image
    unsharp_size : int
        half window size in pixels for unsharp mask
    method : str (either "convolve" or "median")
        method for applying the unsharp mask
    alpha : float
        threshold for deciding a pixel is a cosmic ray spike, i.e. difference between unsharp and smoothed image
    dilation : int
        how many times to dilate pixels identified as spikes, allows for identifying a larger spike region

    Returns
    -------
    (np.ndarray, np.ndarray)
        an image with spikes replaced by the average of their neighbors and the locations of all spikes

    """
    image = image.copy()  # copy to avoid mutating the existing data
    # compute the sizes and smoothing kernel to be used
    kernel_size = unsharp_size * 2 + 1
    smoothing_size = kernel_size * 2 + 1
    smoothing_kernel = radial_array((smoothing_size, smoothing_size)) <= (smoothing_size / 2)
    smoothing_kernel = smoothing_kernel / np.sum(smoothing_kernel)

    # depending on the method, perform unsharping
    if method == "median":
        normalized_image = image / medfilt2d(image, kernel_size=smoothing_size)
        unsharped_image = normalized_image - medfilt2d(normalized_image, kernel_size=kernel_size) > alpha
    elif method == "convolve":
        normalized_image = image / convolve2d(image, smoothing_kernel, mode="same")
        unsharp_kernel = -np.ones((kernel_size, kernel_size)) / kernel_size / kernel_size
        unsharp_kernel[kernel_size // 2, kernel_size // 2] += 0.75
        unsharp_kernel[kernel_size // 2 - 1: kernel_size // 2 + 1, kernel_size // 2 - 1: kernel_size // 2 + 1] += (
            0.25 / 9
        )
        unsharped_image = convolve2d(normalized_image, unsharp_kernel, mode="same") > alpha
    else:
        msg = f"Unsupported method. Method must be 'median' or 'convolve' but received {method}"
        raise NotImplementedError(msg)

    # optional dilation
    if dilation != 0:
        dilation_size = 2 * dilation + 1
        unsharped_image = convolve2d(unsharped_image, np.ones((dilation_size, dilation_size)), mode="same") != 0

    # detect the spikes and fill them with their neighbors
    spikes = np.where(unsharped_image != 0)
    output = np.copy(image)
    image[spikes] = np.nan
    for x, y in zip(*spikes, strict=False):
        neighbors = cell_neighbors(image, x, y, kernel_size - 1)
        threshold = np.nanmedian(neighbors) + 3 * np.nanstd(neighbors)
        output[x, y] = np.nanmean(neighbors[neighbors < threshold])

    return output, spikes


@punch_task
def despike_task(data_object: NDCube,
                 unsharp_size: int = 3,
                 method: str = "convolve",
                 alpha: float = 1,
                 dilation: int = 0) -> NDCube:
    """
    Prefect task to perform despiking.

    Parameters
    ----------
    data_object : NDCube
        data to operate on
    unsharp_size : int
        half window size in pixels for unsharp mask
    method : str (either "convolve" or "median")
        method for applying the unsharp mask
    alpha : float
        threshold for deciding a pixel is a cosmic ray spike, i.e. difference between unsharp and smoothed image
    dilation : int
        how many times to dilate pixels identified as spikes, allows for identifying a larger spike region

    Returns
    -------
    NDCube
        a modified version of the input with spikes removed

    """
    data_object.data[...], spikes = spikejones(
        data_object.data[...], unsharp_size=unsharp_size, method=method, alpha=alpha, dilation=dilation,
    )
    data_object.uncertainty.array[spikes] = np.inf
    data_object.meta.history.add_now("LEVEL1-despike", "image despiked")
    data_object.meta.history.add_now("LEVEL1-despike", f"method={method}")
    data_object.meta.history.add_now("LEVEL1-despike", f"unsharp_size={unsharp_size}")
    data_object.meta.history.add_now("LEVEL1-despike", f"alpha={alpha}")
    data_object.meta.history.add_now("LEVEL1-despike", f"dilation={dilation}")

    return data_object



import numpy as np
from matplotlib.colors import LinearSegmentedColormap

from punchbowl.data.visualize import cmap_punch, radial_distance, radial_filter


def test_radial_distance():
    shape = (100, 100)
    radial_array = radial_distance(shape[0], shape[1])
    assert isinstance(radial_array, np.ndarray)
    assert radial_array.shape == shape


def test_radial_filter():
    shape = (100, 100)
    data_array = np.random.random(shape)
    filtered_array = radial_filter(data_array)
    assert isinstance(filtered_array, np.ndarray)
    assert filtered_array.shape == shape


def test_cmap_punch():
    assert isinstance(cmap_punch, LinearSegmentedColormap)



import numpy as np
from ndcube import NDCube

from punchbowl.data.units import split_ccd_array
from punchbowl.prefect import punch_task


def dn_to_photons(data_array: np.ndarray, gain_left: float = 4.9, gain_right: float = 4.9) -> np.ndarray:
    """Convert an input array from DN to photon count."""
    gain = split_ccd_array(data_array.shape, gain_left, gain_right)
    return data_array * gain


def compute_noise(
        data: np.ndarray,
        bias_level: float = 100,
        dark_level: float = 55.81,
        gain_left: float = 4.9,
        gain_right: float = 4.9,
        read_noise_level: float = 17,
        bitrate_signal: int = 16) -> tuple[np.ndarray, np.ndarray]:
    """
    Generate noise based on an input data array, with specified noise parameters.

    Parameters
    ----------
    data
        input data array (n x n)
    bias_level
        ccd bias level
    dark_level
        ccd dark level
    gain_left
        ccd gain (left side of CCD)
    gain_right
        ccd gain (right side of CCD)
    read_noise_level
        ccd read noise level
    bitrate_signal
        desired ccd data bit level

    Returns
    -------
    data : np.ndarray
        clipped data with the bias level added
    noise : np.ndarray
        computed noise array corresponding to input data and ccd/noise parameters

    """
    data = data.copy().astype("long")

    gain = split_ccd_array(data.shape, gain_left, gain_right)

    # Photon / shot noise generation
    data_photon = data * gain  # DN to photoelectrons
    sigma_photon = np.sqrt(data_photon)  # Converting sigma of this
    sigma = sigma_photon / gain  # Converting back to DN
    noise_photon = np.random.normal(scale=sigma)

    # Add bias level and clip pixels to avoid overflow
    data = np.clip(data + bias_level, 0, 2 ** bitrate_signal - 1)

    # Dark noise generation
    noise_level = dark_level * gain
    noise_dark = np.random.poisson(lam=noise_level, size=data.shape) / gain

    # Read noise generation
    noise_read = np.random.normal(scale=read_noise_level, size=data.shape)
    noise_read = noise_read / gain  # Convert back to DN

    # And then add noise terms directly
    return data, noise_photon + noise_dark + noise_read


def compute_uncertainty(data_array: np.ndarray,
                        bias_level: float = 100,
                        dark_level: float = 55.81,
                        gain_left: float = 4.9,
                        gain_right: float = 4.9,
                        read_noise_level: float = 17,
                        bitrate_signal: int = 16,
                        ) -> np.ndarray:
    """With an input data array compute a corresponding uncertainty array."""
    # Convert this photon count to a shot noise
    _, noise = compute_noise(data_array,
                                bias_level=bias_level,
                                dark_level=dark_level,
                                gain_left=gain_left,
                                gain_right=gain_right,
                                read_noise_level=read_noise_level,
                                bitrate_signal=bitrate_signal)
    return noise


@punch_task
def update_initial_uncertainty_task(data_object: NDCube,
                                    bias_level: float = 100,
                                    dark_level: float = 55.81,
                                    gain_left: float = 4.9,
                                    gain_right: float = 4.9,
                                    read_noise_level: float = 17,
                                    bitrate_signal: int = 16) -> NDCube:
    """Prefect task to compute initial uncertainty."""
    uncertainty_array = compute_uncertainty(data_object.data,
                                            bias_level=bias_level,
                                            dark_level=dark_level,
                                            gain_left=gain_left,
                                            gain_right=gain_right,
                                            read_noise_level=read_noise_level,
                                            bitrate_signal=bitrate_signal,
                                            )
    data_object.uncertainty.array = uncertainty_array

    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", "Initial uncertainty computed with:")
    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", f"bias_level={bias_level}")
    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", f"dark_level={dark_level}")
    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", f"gain_left={gain_left}")
    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", f"gain_right={gain_right}")
    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", f"read_noise_level={read_noise_level}")
    data_object.meta.history.add_now("LEVEL1-initial_uncertainty", f"bitrate_signal={bitrate_signal}")
    return data_object



from datetime import datetime

import astropy.units as u
import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube
from prefect import flow, get_run_logger
from remove_starfield import Starfield

from punchbowl.data import NormalizedMetadata
from punchbowl.data.wcs import calculate_celestial_wcs_from_helio, calculate_helio_wcs_from_celestial, get_p_angle
from punchbowl.level3.stellar import (
    from_celestial,
    generate_starfield_background,
    subtract_starfield_background_task,
    to_celestial,
)


@pytest.fixture()
def one_data(shape: tuple = (128, 128)) -> NDCube:
    """
    Generate some random data for testing
    """
    data = np.ones(shape)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 64, 64
    wcs.wcs.crval = 0, 24.75

    meta = NormalizedMetadata(
        {"TYPECODE": "CS", "OBSCODE": "M", "LEVEL": "3", "OBSRVTRY": "0", "DATE-OBS": "2024-04-08T18:40:00"})
    return NDCube(data=data, wcs=wcs, meta=meta, uncertainty=StdDevUncertainty(np.zeros_like(data)))


@pytest.fixture()
def zero_starfield_data(shape: tuple = (256, 256)) -> Starfield:
    """
    Generate some random data for testing
    """
    starfield = np.zeros(shape)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 128, 128
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    return Starfield(starfield=starfield, wcs=wcs)


@pytest.fixture
def sample_ndcube() -> NDCube:
    def _sample_ndcube(shape: tuple, code: str = "PTM", level: str = "2") -> NDCube:
        data = np.ones(shape).astype(np.float32)
        sqrt_abs_data = np.sqrt(np.abs(data))
        uncertainty = StdDevUncertainty(np.interp(sqrt_abs_data, (sqrt_abs_data.min(), sqrt_abs_data.max()),
                                                  (0, 1)).astype(np.float32))
        wcs = WCS(naxis=3)
        wcs.ctype = "WAVE", "HPLT-TAN", "HPLN-TAN"
        wcs.cdelt = 0.2, 0.1, 0.1
        wcs.cunit = "Angstrom", "deg", "deg"
        wcs.crpix = 0, 0, 0
        wcs.crval = 5, 1, 1

        meta = NormalizedMetadata.load_template(code, level)
        meta["DATE-OBS"] = str(datetime(2024, 3, 21, 00, 00, 00))
        meta["FILEVRSN"] = "1"
        return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)

    return _sample_ndcube


def test_from_celestial(sample_ndcube) -> NDCube:
    test_cube = sample_ndcube((3, 10, 10))
    data_solar = from_celestial(test_cube)
    expected_solar = np.full((3, 10, 10), 1.0, dtype=np.float32)
    assert isinstance(data_solar, NDCube)
    assert np.allclose(data_solar.data, expected_solar)

def test_to_celestial(sample_ndcube) -> NDCube:
    test_cube = sample_ndcube((3, 10, 10))
    data_cel = to_celestial(test_cube)
    expected_cel = np.full((3, 10, 10), 1.0, dtype=np.float32)
    assert isinstance(data_cel, NDCube)
    assert np.allclose(data_cel.data, expected_cel)



import os

import numpy as np
from ndcube import NDCube
from numpy.lib.stride_tricks import as_strided

from punchbowl.data import load_ndcube_from_fits
from punchbowl.prefect import punch_task


def sliding_window(arr: np.ndarray, window_size: int) -> np.ndarray:
    """
    Construct a sliding window view of the array.

    borrowed from: https://stackoverflow.com/questions/10996769/pixel-neighbors-in-2d-array-image-using-python.
    """
    arr = np.asarray(arr)
    window_size = int(window_size)
    if arr.ndim != 2:
        msg = "need 2-D input"
        raise ValueError(msg)
    if not (window_size > 0):
        msg = "need a positive window size"
        raise ValueError(msg)
    shape = (arr.shape[0] - window_size + 1, arr.shape[1] - window_size + 1, window_size, window_size)
    if shape[0] <= 0:
        shape = (1, shape[1], arr.shape[0], shape[3])
    if shape[1] <= 0:
        shape = (shape[0], 1, shape[2], arr.shape[1])
    strides = (arr.shape[1] * arr.itemsize, arr.itemsize, arr.shape[1] * arr.itemsize, arr.itemsize)
    return as_strided(arr, shape=shape, strides=strides)


def cell_neighbors(arr: np.ndarray, i: int, j: int, window_size: int = 1) -> np.ndarray:
    """
    Return d-th neighbors of cell (i, j).

    borrowed from: https://stackoverflow.com/questions/10996769/pixel-neighbors-in-2d-array-image-using-python.
    """
    window = sliding_window(arr, 2 * window_size + 1)

    ix = np.clip(i - window_size, 0, window.shape[0] - 1)
    jx = np.clip(j - window_size, 0, window.shape[1] - 1)

    i0 = max(0, i - window_size - ix)
    j0 = max(0, j - window_size - jx)
    i1 = window.shape[2] - max(0, window_size - i + ix)
    j1 = window.shape[3] - max(0, window_size - j + jx)

    return window[ix, jx][i0:i1, j0:j1].ravel()


def mean_correct(
    data_array: np.ndarray, mask_array: np.ndarray, required_good_count: int = 3, max_window_size: int = 10,
) -> np.ndarray:
    """Mean correct."""
    x_bad_pix, y_bad_pix = np.where(mask_array == 0)
    data_array[mask_array == 0] = 0
    output_data_array = data_array.copy()
    for x_i, y_i in zip(x_bad_pix, y_bad_pix, strict=False):
        window_size = 1
        number_good_px = np.sum(cell_neighbors(mask_array, x_i, y_i, window_size=window_size))
        while number_good_px < required_good_count:
            window_size += 1
            number_good_px = np.sum(cell_neighbors(mask_array, x_i, y_i, window_size=window_size))
            if window_size > max_window_size:
                break
        output_data_array[x_i, y_i] = (
            np.sum(cell_neighbors(data_array, x_i, y_i, window_size=window_size)) / number_good_px
        )

    return output_data_array


def median_correct(
    data_array: np.ndarray, mask_array: np.ndarray, required_good_count: int = 3, max_window_size: int = 10,
) -> np.ndarray:
    """Median correct."""
    x_bad_pix, y_bad_pix = np.where(mask_array == 0)
    data_array[mask_array == 0] = np.nan
    output_data_array = data_array.copy()
    for x_i, y_i in zip(x_bad_pix, y_bad_pix, strict=False):
        window_size = 1
        number_good_px = np.sum(cell_neighbors(mask_array, x_i, y_i, window_size=window_size))
        while number_good_px < required_good_count:
            window_size += 1
            number_good_px = np.sum(cell_neighbors(mask_array, x_i, y_i, window_size=window_size))
            if window_size > max_window_size:
                break
        output_data_array[x_i, y_i] = np.nanmedian(cell_neighbors(data_array, x_i, y_i, window_size=window_size))

    return output_data_array


def remove_deficient_pixels(data: NDCube,
                            deficient_pixels: np.ndarray,
                            required_good_count: int = 3,
                            max_window_size: int = 10,
                            method: str = "median") -> NDCube:
    """Remove deficient pixels."""
    # check dimensions match
    if data.data.shape != deficient_pixels.shape:
        msg = (
            "deficient_pixel_array expects the data_object and"
            "deficient_pixel_array arrays to have the same dimensions."
            f"data_array dims: {data.data.shape}"
            f"and deficient_pixel_map dims: {deficient_pixels.shape}"
        )
        raise ValueError(
            msg,
        )

    if np.all(deficient_pixels == 0):
        msg = "All pixels in mask are deficient and cannot be corrected."
        raise ValueError(msg)

    if method == "median":
        data_array = median_correct(
            data.data, deficient_pixels, required_good_count=required_good_count, max_window_size=max_window_size,
        )
    elif method == "mean":
        data_array = mean_correct(
            data.data, deficient_pixels, required_good_count=required_good_count, max_window_size=max_window_size,
        )
    else:
        msg = f"method specified must be 'mean', or 'median'. Found method={method}"
        raise ValueError(msg)

    # Set deficient pixels to complete uncertainty
    output_uncertainty = data.uncertainty.array.copy()
    output_uncertainty[deficient_pixels == 0] = np.inf

    data.data[...] = data_array[...]
    data.uncertainty.array = output_uncertainty
    return data


@punch_task
def remove_deficient_pixels_task(
    data: NDCube,
    deficient_pixel_map_path: str | None,
    required_good_count: int = 3,
    max_window_size: int = 10,
    method: str = "median",
) -> NDCube:
    """
    Subtracts a deficient pixel map from an input data frame.

    checks the dimensions of input data frame and map match and
    subtracts the background model from the data frame of interest.

    Parameters
    ----------
    data : NDCube
        A PUNCHobject data frame to be background subtracted

    deficient_pixel_map_path : Optional[str]
        The path to the deficient pixel map use to in correction

    required_good_count : int
        how many neighboring pixels must not be deficient to correct a pixel,
            if fewer than that many pixels are good neighbors then the box expands

    max_window_size : int
        the width of the max window

    method : str
        either "mean" or "median" depending on which measure should fill the deficient pixel


    Returns
    -------
    NDCube
        A background subtracted data frame

    """
    if deficient_pixel_map_path is None:
        output_object = data
        output_object.meta.history.add_now("LEVEL1-remove_deficient_pixels",
                                           "Remove deficient pixels skipped since path is empty")

    else:
        deficient_pixel_map = load_ndcube_from_fits(deficient_pixel_map_path)

        deficient_pixel_array = deficient_pixel_map.data
        output_object = remove_deficient_pixels(data, deficient_pixel_array,
                                                required_good_count=required_good_count,
                                                max_window_size=max_window_size,
                                                method=method)
        output_object.meta["CALPM"] = os.path.basename(deficient_pixel_map_path)
        output_object.meta.history.add_now("LEVEL1-remove_deficient_pixels", "deficient pixels removed")

    return output_object


def create_all_valid_deficient_pixel_map(data: NDCube) -> NDCube:
    """Create valid deficient pixel map."""
    data.data[...] = np.ones_like(data.data)
    return data



import os

import numpy as np
import pytest
import reproject
from astropy.io import fits

from punchbowl.data import punch_io
from punchbowl.level3 import celestial_intermediary

TESTDATA_DIR = os.path.dirname(__file__)
TEST_FILE = TESTDATA_DIR + '/data/downsampled_L2_PTM.fits'


def test_to_celestial_frame_cutout():
    data_cube = punch_io.load_ndcube_from_fits(TEST_FILE, include_provenance=False)
    reprojected_cube = celestial_intermediary.to_celestial_frame_cutout(data_cube, cdelt=1)
    assert np.any(np.isfinite(reprojected_cube.data))
    assert np.any(np.isfinite(reprojected_cube.uncertainty.array))

    assert 0 <= reprojected_cube.wcs.wcs.crval[0] < 360
    assert reprojected_cube.wcs.wcs.crval[1] == 0
    assert reprojected_cube.wcs.wcs.cdelt[0] == reprojected_cube.wcs.wcs.cdelt[1]
    assert tuple(reprojected_cube.wcs.wcs.ctype) == ('RA---CAR', 'DEC--CAR')


# A point of concern is whether this shifting of reprojected images works across the RA=0 wrap point, so we test that
# well here.
@pytest.mark.parametrize('shift1,shift2,is_overlap',
                          # This is a "benign" pair, nowhere near the RA=0 point
                         [((0, 0), (-12, 5), True),
                          # This pair is right on either side of RA=0, but neither contains that point
                          ((40, 12), (-150, 0), False),
                          # The first one is at high RA and the second one straddles RA=0
                          ((40, 12), (-230, 0), True),
                          # Now one at low RA and one straddling RA=0
                          ((-150, 0), (-230, 1), True),
                         ])
def test_shift_image_onto(tmp_path, shift1, shift2, is_overlap):
    # Generate a test file, where we customize which portion of the sky it covers
    file1 = tmp_path / 'shifted_file1.fits'
    with fits.open(TEST_FILE) as hdul:
        for hdu in hdul[1:]:
            hdu.header['CRVAL1A'] += shift1[0]
            hdu.header['CRVAL2A'] += shift1[1]
            # Cut down to one Stokes component to speed up the test
            hdu.data = hdu.data[0]
        hdul.writeto(file1)

    # Generate another test file, which covers a different portion of the sky
    file2 = tmp_path / 'shifted_file2.fits'
    with fits.open(TEST_FILE) as hdul:
        for hdu in hdul[1:]:
            hdu.header['CRVAL1A'] += shift2[0]
            hdu.header['CRVAL2A'] += shift2[1]
            # Cut down to one Stokes component to speed up the test
            hdu.data = hdu.data[0]
        hdul.writeto(file2)

    cube1 = punch_io.load_ndcube_from_fits(file1, key='A', include_provenance=False)
    cube2 = punch_io.load_ndcube_from_fits(file2, key='A', include_provenance=False)
    reproj_data_1 = celestial_intermediary.to_celestial_frame_cutout(cube1, cdelt=.6)
    reproj_data_2 = celestial_intermediary.to_celestial_frame_cutout(cube2, cdelt=.6)

    # Just sanity-check that these are different images and different outputs, so we know the rest of the test is valid
    assert reproj_data_1.wcs.wcs.crval[0] != reproj_data_2.wcs.wcs.crval[0]
    assert reproj_data_1.data.shape != reproj_data_2.data.shape or not np.allclose(
        reproj_data_1.data, reproj_data_2.data, equal_nan=True)

    # And sanity-check that the data came out valid
    assert np.any(np.isfinite(reproj_data_1.data))
    assert np.any(np.isfinite(reproj_data_1.uncertainty.array))
    assert np.any(np.isfinite(reproj_data_2.data))
    assert np.any(np.isfinite(reproj_data_2.uncertainty.array))

    # The celestial intermediary images should be different cutouts of the sky, but on the same pixel grid. We test
    # that by aligning them first with this function, which simply adds and removes rows of pixels, and then by
    # actually reprojecting the data into the same output frame.
    d1_shifted_to_2 = celestial_intermediary.shift_image_onto(reproj_data_1, reproj_data_2)
    assert not is_overlap or np.any(np.isfinite(d1_shifted_to_2.data))
    assert not is_overlap or np.any(np.isfinite(d1_shifted_to_2.uncertainty.array))

    data = np.stack((cube1.data, cube1.uncertainty.array))
    d1_reprojected_to_2 = reproject.reproject_adaptive((data, cube1.wcs.celestial), reproj_data_2.wcs,
                                                       reproj_data_2.data.shape, return_footprint=False,
                                                       roundtrip_coords=False, boundary_mode='strict',
                                                       conserve_flux=True, center_jacobian=True)

    # We should be getting the same thing, to a part in 10^8
    np.testing.assert_allclose(d1_shifted_to_2.data, d1_reprojected_to_2[0], equal_nan=True, rtol=1e-8)
    np.testing.assert_allclose(d1_shifted_to_2.uncertainty.array, d1_reprojected_to_2[1], equal_nan=True, rtol=1e-8)

    # Now do it the other way
    d2_shifted_to_1 = celestial_intermediary.shift_image_onto(reproj_data_2, reproj_data_1)
    assert not is_overlap or np.any(np.isfinite(d2_shifted_to_1.data))

    data = np.stack((cube2.data, cube2.uncertainty.array))
    d2_reprojected_to_1 = reproject.reproject_adaptive((data, cube2.wcs.celestial), reproj_data_1.wcs,
                                                       reproj_data_1.data.shape, return_footprint=False,
                                                       roundtrip_coords=False, boundary_mode='strict',
                                                       conserve_flux=True, center_jacobian=True)

    # We should be getting the same thing, to a part in 10^8
    np.testing.assert_allclose(d2_shifted_to_1.data, d2_reprojected_to_1[0], equal_nan=True, rtol=1e-8)
    np.testing.assert_allclose(d2_shifted_to_1.uncertainty.array, d2_reprojected_to_1[1], equal_nan=True, rtol=1e-8)


def test_shift_image_onto_3d_cube(tmp_path):
    # Generate a test file, where we customize which portion of the sky it covers
    file1 = TEST_FILE

    # Generate another test file, which covers a different portion of the sky
    file2 = tmp_path / 'shifted_file2.fits'
    with fits.open(TEST_FILE) as hdul:
        for hdu in hdul[1:]:
            hdu.header['CRVAL1A'] += 10
            hdu.header['CRVAL2A'] += 5
        hdul.writeto(file2)

    cube1 = punch_io.load_ndcube_from_fits(file1, key='A', include_provenance=False)
    cube2 = punch_io.load_ndcube_from_fits(file2, key='A', include_provenance=False)
    reproj_data_1 = celestial_intermediary.to_celestial_frame_cutout(cube1, cdelt=.6)
    reproj_data_2 = celestial_intermediary.to_celestial_frame_cutout(cube2, cdelt=.6)

    # Just sanity-check that these are different images and different outputs, so we know the rest of the test is valid
    assert reproj_data_1.wcs.wcs.crval[0] != reproj_data_2.wcs.wcs.crval[0]
    assert reproj_data_1.data.shape != reproj_data_2.data.shape or not np.allclose(
        reproj_data_1, reproj_data_2, equal_nan=True)

    # And sanity-check that the data came out valid
    assert np.any(np.isfinite(reproj_data_1.data))
    assert np.any(np.isfinite(reproj_data_1.uncertainty.array))
    assert np.any(np.isfinite(reproj_data_2.data))
    assert np.any(np.isfinite(reproj_data_2.uncertainty.array))

    # The celestial intermediary images should be different cutouts of the sky, but on the same pixel grid. We test
    # that by aligning them first with this function, which simply adds and removes rows of pixels, and then by
    # actually reprojecting the data into the same output frame.
    d1_shifted_to_2 = celestial_intermediary.shift_image_onto(reproj_data_1, reproj_data_2)
    assert np.any(np.isfinite(d1_shifted_to_2.data))
    assert np.any(np.isfinite(d1_shifted_to_2.uncertainty.array))

    data = np.stack((cube1.data, cube1.uncertainty.array))
    d1_reprojected_to_2 = reproject.reproject_adaptive((data, cube1.wcs.celestial), reproj_data_2.wcs,
                                                       reproj_data_2.data.shape[1:], return_footprint=False,
                                                       roundtrip_coords=False, boundary_mode='strict',
                                                       conserve_flux=True, center_jacobian=True)

    # We should be getting the same thing, to a part in 10^8
    np.testing.assert_allclose(d1_shifted_to_2.data, d1_reprojected_to_2[0], equal_nan=True, rtol=1e-8)
    np.testing.assert_allclose(d1_shifted_to_2.uncertainty.array, d1_reprojected_to_2[1], equal_nan=True, rtol=1e-8)

    # Now do it the other way
    d2_shifted_to_1 = celestial_intermediary.shift_image_onto(reproj_data_2, reproj_data_1)
    assert np.any(np.isfinite(d2_shifted_to_1.data))

    data = np.stack((cube2.data, cube2.uncertainty.array))
    d2_reprojected_to_1 = reproject.reproject_adaptive((data, cube2.wcs.celestial), reproj_data_1.wcs,
                                                       reproj_data_1.data.shape[1:], return_footprint=False,
                                                       roundtrip_coords=False, boundary_mode='strict',
                                                       conserve_flux=True, center_jacobian=True)

    # We should be getting the same thing, to a part in 10^8
    np.testing.assert_allclose(d2_shifted_to_1.data, d2_reprojected_to_1[0], equal_nan=True, rtol=1e-8)
    np.testing.assert_allclose(d2_shifted_to_1.uncertainty.array, d2_reprojected_to_1[1], equal_nan=True, rtol=1e-8)


def test_shift_image_onto_fill_value(tmp_path):
    file1 = TEST_FILE

    # Generate another test file, which covers a different portion of the sky
    file2 = tmp_path / 'shifted_file2.fits'
    with fits.open(TEST_FILE) as hdul:
        for hdu in hdul[1:]:
            hdu.header['CRVAL1A'] += 10
            hdu.header['CRVAL2A'] += 10
        hdul.writeto(file2)

    cube1 = punch_io.load_ndcube_from_fits(file1, key='A', include_provenance=False)
    cube2 = punch_io.load_ndcube_from_fits(file2, key='A', include_provenance=False)
    reproj_data_1 = celestial_intermediary.to_celestial_frame_cutout(cube1, cdelt=.6)
    reproj_data_2 = celestial_intermediary.to_celestial_frame_cutout(cube2, cdelt=.6)

    d1_shifted_to_2 = celestial_intermediary.shift_image_onto(reproj_data_1, reproj_data_2, fill_value=np.nan)
    d2_shifted_to_1 = celestial_intermediary.shift_image_onto(reproj_data_2, reproj_data_1, fill_value=np.nan)

    assert np.any(np.isnan(d1_shifted_to_2.data))
    assert np.any(np.isnan(d2_shifted_to_1.data))

    # Make sure there are no NaNs in the input
    reproj_data_1.data[...] = np.nan_to_num(reproj_data_1.data)
    reproj_data_2.data[...] = np.nan_to_num(reproj_data_2.data)
    reproj_data_1.uncertainty.array[...] = np.nan_to_num(reproj_data_1.uncertainty.array)
    reproj_data_2.uncertainty.array[...] = np.nan_to_num(reproj_data_2.uncertainty.array)

    d1_shifted_to_2 = celestial_intermediary.shift_image_onto(reproj_data_1, reproj_data_2, fill_value=0)
    d2_shifted_to_1 = celestial_intermediary.shift_image_onto(reproj_data_2, reproj_data_1, fill_value=0)

    assert not np.any(np.isnan(d1_shifted_to_2.data))
    assert not np.any(np.isnan(d2_shifted_to_1.data))
    assert not np.any(np.isnan(d1_shifted_to_2.uncertainty.array))
    assert not np.any(np.isnan(d2_shifted_to_1.uncertainty.array))


def test_shift_image_onto_different_cdelts():
    reproj_data_1 = celestial_intermediary.to_celestial_frame_cutout(
        punch_io.load_ndcube_from_fits(TEST_FILE, key='A', include_provenance=False), cdelt=1.2)
    reproj_data_2 = celestial_intermediary.to_celestial_frame_cutout(
        punch_io.load_ndcube_from_fits(TEST_FILE, key='A', include_provenance=False), cdelt=1)

    with pytest.raises(ValueError, match=".*WCSes must have identical CDELTs.*"):
        celestial_intermediary.shift_image_onto(reproj_data_1, reproj_data_2)



import os
from datetime import UTC, datetime

import numpy as np
import pytest
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS, DistortionLookupTable
from astropy.wcs.utils import add_stokes_axis_to_wcs
from ndcube import NDCube

from punchbowl.data.meta import NormalizedMetadata
from punchbowl.data.punch_io import (
    CALIBRATION_ANNOTATION,
    _update_statistics,
    get_base_file_name,
    load_ndcube_from_fits,
    write_ndcube_to_fits,
    write_ndcube_to_quicklook,
)

TESTDATA_DIR = os.path.dirname(__file__)
SAMPLE_FITS_PATH_UNCOMPRESSED = os.path.join(TESTDATA_DIR, "test_data.fits")
SAMPLE_FITS_PATH_COMPRESSED = os.path.join(TESTDATA_DIR, "test_data.fits")
SAMPLE_OMNIBUS_PATH = os.path.join(TESTDATA_DIR, "omniheader.csv")
SAMPLE_LEVEL_PATH = os.path.join(TESTDATA_DIR, "LevelTest.yaml")
SAMPLE_SPACECRAFT_DEF_PATH = os.path.join(TESTDATA_DIR, "spacecraft.yaml")


@pytest.fixture
def sample_ndcube():
    def _sample_ndcube(shape, code="PM1", level="0"):
        data = np.random.random(shape).astype(np.float32)
        uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))
        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
        wcs.wcs.cunit = "deg", "deg"
        wcs.wcs.cdelt = 0.1, 0.1
        wcs.wcs.crpix = 0, 0
        wcs.wcs.crval = 1, 1
        wcs.wcs.cname = "HPC lon", "HPC lat"

        if level in ["2", "3"] and code[0] == "P":
            wcs = add_stokes_axis_to_wcs(wcs, 2)

        meta = NormalizedMetadata.load_template(code, level)
        meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1))
        meta['FILEVRSN'] = "1"
        return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)
    return _sample_ndcube


def test_write_data(sample_ndcube, tmpdir):
    cube = sample_ndcube((50, 50))
    cube.meta["LEVEL"] = "1"
    cube.meta["TYPECODE"] = "CL"
    cube.meta["OBSRVTRY"] = "1"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now(UTC))
    cube.meta["DATE-END"] = str(datetime.now(UTC))

    test_path = os.path.join(tmpdir, "test.fits")
    write_ndcube_to_fits(cube, test_path)
    assert os.path.isfile(test_path)

    with fits.open(test_path) as hdul:
        assert hdul[1].header['EXTNAME'] == "PRIMARY DATA ARRAY"
        assert hdul[2].header['EXTNAME'] == "UNCERTAINTY ARRAY"
        assert hdul[3].header['EXTNAME'] == "FILE PROVENANCE"


def test_write_data_jp2(sample_ndcube, tmpdir):
    cube = sample_ndcube((50, 50))
    cube.meta["LEVEL"] = "1"
    cube.meta["TYPECODE"] = "CL"
    cube.meta["OBSRVTRY"] = "1"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now(UTC))
    cube.meta["DATE-END"] = str(datetime.now(UTC))

    test_path = os.path.join(tmpdir, "test.jp2")
    write_ndcube_to_quicklook(cube, test_path)
    assert os.path.isfile(test_path)

def test_write_jpeg(sample_ndcube, tmpdir):
    from punchbowl.data.sample import PUNCH_PAM
    cube = sample_ndcube((4096, 4096))
    with fits.open(PUNCH_PAM) as hdul:
        cube.data[...] = hdul[1].data[0]
    cube.meta["LEVEL"] = "1"
    cube.meta["TYPECODE"] = "CL"
    cube.meta["OBSRVTRY"] = "1"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now(UTC))
    cube.meta["DATE-END"] = str(datetime.now(UTC))

    test_path = os.path.join(tmpdir, "test.jpeg")
    write_ndcube_to_quicklook(cube, test_path, vmin=1E-15,  vmax=8E-12, annotation="Hi there! I'm {TYPECODE}.", color=True)
    assert os.path.isfile(test_path)


def test_write_data_jp2_with_annotation(sample_ndcube, tmpdir):
    cube = sample_ndcube((2048, 2048))
    cube.meta["LEVEL"] = "1"
    cube.meta["TYPECODE"] = "CL"
    cube.meta["OBSRVTRY"] = "1"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now(UTC))
    cube.meta["DATE-END"] = str(datetime.now(UTC))

    test_path = os.path.join(tmpdir, "test.jp2")
    write_ndcube_to_quicklook(cube, test_path, annotation=CALIBRATION_ANNOTATION)
    assert os.path.isfile(test_path)


def test_write_data_jp2_wrong_filename(sample_ndcube, tmpdir):
    cube = sample_ndcube((50, 50))
    cube.meta["LEVEL"] = "1"
    cube.meta["TYPECODE"] = "CL"
    cube.meta["OBSRVTRY"] = "1"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now(UTC))
    cube.meta["DATE-END"] = str(datetime.now(UTC))

    test_path = os.path.join(tmpdir, "test.fits")
    with pytest.raises(ValueError):
        write_ndcube_to_quicklook(cube, test_path)


def test_write_data_jp2_wrong_dimensions(sample_ndcube, tmpdir):
    cube = sample_ndcube((2, 50, 50))
    cube.meta["LEVEL"] = "3"
    cube.meta["TYPECODE"] = "PAM"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now(UTC))
    cube.meta["DATE-END"] = str(datetime.now(UTC))

    test_path = os.path.join(tmpdir, "test.jp2")
    with pytest.raises(ValueError):
        write_ndcube_to_quicklook(cube, test_path)


def test_generate_data_statistics_from_zeros():
    w = WCS(naxis=2)
    m = NormalizedMetadata.load_template("PM1", "0")
    m.history.add_now("Test", "does it write?")
    m.history.add_now("Test", "how about twice?")
    m['DESCRPTN'] = 'This is a test!'
    m['CHECKSUM'] = ''
    m['DATASUM'] = ''
    m.delete_section("World Coordinate System")

    sample_data = NDCube(data=np.zeros((2048, 2048),dtype=np.int16), wcs=w, meta=m)

    _update_statistics(sample_data)

    assert sample_data.meta['DATAZER'].value == 2048*2048

    assert sample_data.meta['DATASAT'].value == 0

    assert sample_data.meta['DATAAVG'].value == -999.0
    assert sample_data.meta['DATAMDN'].value == -999.0
    assert sample_data.meta['DATASIG'].value == -999.0

    percentile_percentages = [1, 10, 25, 50, 75, 90, 95, 98, 99]
    percentile_values = [-999.0 for _ in percentile_percentages]

    for percent, value in zip(percentile_percentages, percentile_values):
        assert sample_data.meta[f'DATAP{percent:02d}'].value == value

    assert sample_data.meta['DATAMIN'].value == 0.0
    assert sample_data.meta['DATAMAX'].value == 0.0


def test_generate_data_statistics(sample_ndcube):
    cube = sample_ndcube((50, 50))
    _update_statistics(cube)

    nonzero_sample_data = cube.data[np.where(cube.data != 0)].flatten()

    assert cube.meta['DATAZER'].value == len(np.where(cube.data == 0)[0])

    assert cube.meta['DATASAT'].value == len(np.where(cube.data >= cube.meta['DSATVAL'].value)[0])

    assert cube.meta['DATAAVG'].value == np.mean(nonzero_sample_data)
    assert cube.meta['DATAMDN'].value == np.median(nonzero_sample_data)
    assert cube.meta['DATASIG'].value == np.std(nonzero_sample_data)

    percentile_percentages = [1, 10, 25, 50, 75, 90, 95, 98, 99]
    percentile_values = np.percentile(nonzero_sample_data, percentile_percentages)

    for percent, value in zip(percentile_percentages, percentile_values):
        assert cube.meta[f'DATAP{percent:02d}'].value == value

    assert cube.meta['DATAMIN'].value == float(cube.data.min())
    assert cube.meta['DATAMAX'].value == float(cube.data.max())


def test_filename_base_generation(sample_ndcube):
    cube = sample_ndcube((50, 50))
    actual = get_base_file_name(cube)
    expected = "PUNCH_L0_PM1_20240222160001_v1"
    assert actual == expected


def test_has_typecode():
    meta = NormalizedMetadata.load_template("CFM", "3")
    meta["DATE-OBS"] = str(datetime.now(UTC))
    assert "TYPECODE" in meta


def test_load_punchdata_with_history(tmpdir):
    data = np.ones((10, 10), dtype=np.uint16)
    meta = NormalizedMetadata.load_template("CR4", "0")
    meta['DATE-OBS'] = str(datetime.now(UTC))
    meta.history.add_now("test", "this is a test!")
    meta.history.add_now("test", "this is a second test!")
    wcs = WCS({"CRVAL1": 0.0,
                     "CRVAL2": 0.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC"})
    obj = NDCube(data=data, wcs=wcs, meta=meta)

    assert "OBSCODE" in obj.meta.fits_keys
    file_path = os.path.join(tmpdir, get_base_file_name(obj) + ".fits")
    write_ndcube_to_fits(obj, file_path, overwrite=True)
    reloaded = load_ndcube_from_fits(file_path)
    assert isinstance(reloaded, NDCube)
    assert len(reloaded.meta.history) == 2
    assert reloaded.data.shape == (10, 10)
    assert np.all(reloaded.data == 1)
    os.remove(file_path)


def make_empty_distortion_model(num_bins: int, image: np.ndarray) -> tuple:
    """ Create an empty distortion table

    Parameters
    ----------
    num_bins : int
        number of histogram bins in the distortion model, i.e. the size of the distortion model is (num_bins, num_bins)
    image : np.ndarray
        image to create a distortion model for

    Returns
    -------
    (DistortionLookupTable, DistortionLookupTable)
        x and y distortion models
    """
    # make an initial empty distortion model
    r = np.linspace(0, image.shape[0], num_bins + 1)
    c = np.linspace(0, image.shape[1], num_bins + 1)
    r = (r[1:] + r[:-1]) / 2
    c = (c[1:] + c[:-1]) / 2

    err_px, err_py = r, c
    err_x = np.zeros((num_bins, num_bins))
    err_y = np.zeros((num_bins, num_bins))

    cpdis1 = DistortionLookupTable(
        -err_x.astype(np.float32), (0, 0), (err_px[0], err_py[0]), ((err_px[1] - err_px[0]), (err_py[1] - err_py[0]))
    )
    cpdis2 = DistortionLookupTable(
        -err_y.astype(np.float32), (0, 0), (err_px[0], err_py[0]), ((err_px[1] - err_px[0]), (err_py[1] - err_py[0]))
    )
    return cpdis1, cpdis2


def test_write_punchdata_with_distortion(tmpdir):
    data = np.ones((2048, 2048), dtype=np.uint16)
    meta = NormalizedMetadata.load_template("CR4", "1")
    meta['DATE-OBS'] = str(datetime.now(UTC))
    meta.history.add_now("test", "this is a test!")
    meta.history.add_now("test", "this is a second test!")
    wcs = WCS({"CRVAL1": 0.0,
                     "CRVAL2": 0.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC"})
    cpdis1, cpdis2 = make_empty_distortion_model(100, data)
    wcs.cpdis1 = cpdis1
    wcs.cpdis2 = cpdis2
    obj = NDCube(data=data, wcs=wcs, meta=meta)
    file_path = os.path.join(tmpdir, get_base_file_name(obj) + ".fits")
    write_ndcube_to_fits(obj, file_path, overwrite=True)

    with fits.open(file_path) as hdul:
        assert len(hdul) == 6

    loaded_cube = load_ndcube_from_fits(file_path)
    assert loaded_cube.wcs.has_distortion



"""punchbowl: data reduction and calibration pipeline for PUNCH."""
import importlib.metadata
from importlib.metadata import PackageNotFoundError

try:
  __version__ = importlib.metadata.version("punchbowl")
except PackageNotFoundError:
  __version__ = "0.0.0"



import pytest
from ndcube import NDCube
from prefect.logging import disable_run_logger
from prefect.testing.utilities import prefect_test_harness

from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.level2.polarization import resolve_polarization, resolve_polarization_task


@pytest.fixture
def sample_data_triplet(sample_ndcube):
    """
    Generate a list of sample PUNCH data objects for testing polarization resolving
    """

    polar_angles = [-60, 0, 60]
    sample_data = []

    for angle in polar_angles:
        sample_pd = sample_ndcube(shape=(50, 50))
        sample_pd.meta['POLAR'] = angle
        sample_data.append(sample_pd)

    return sample_data


def test_resolve_polarization(sample_data_triplet):
    """
        Test the resolve polarization function directly
    """

    output_punchdata_list = resolve_polarization(sample_data_triplet)

    assert all(isinstance(output_punchdata, NDCube) for output_punchdata in output_punchdata_list)


@pytest.mark.prefect_test
def test_resolve_polarization_task(sample_data_triplet):
    """
    Test the resolve polarization prefect flow using a test harness
    """

    with disable_run_logger():
        output_punchdata_list = resolve_polarization_task.fn(sample_data_triplet)

    assert all(isinstance(output_punchdata, NDCube) for output_punchdata in output_punchdata_list)




import numpy as np
import reproject
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS, utils
from ndcube import NDCube


def to_celestial_frame_cutout(data_cube: NDCube, cdelt: float = 0.02) -> NDCube:
    """
    Reproject an image to a bounding-box cutout of an all-sky map.

    The bounding box of the input image in the celestial frame is found, and that cutout of the all-sky map is
    returned with the input image reprojected into it. All outputs from this function will be on the same pixel grid
    (if the same ``cdelt`` value is provided), so that different outputs from this function can easily be co-aligned by
    adding and removing pixel rows and columns (no reprojected needed).

    Input files must have two image dimensions, and can have arbitrarily-many extra leading dimensions.

    The all-sky map used is a plate carrée (CAR) projection. The CRPIX/CRVAL values of the WCSes returned by this
    function will vary in longitude so that the output frame is approximately centered on the output data,
    but CRVAL latitude will always be zero (as otherwise the CAR projection no longer has straight lat/lon lines).

    The motivation behind this function is to reduce reprojection time and disk space usage when saving a lot of
    images in celestial coordinates, but still keep the stackability that the output images would have if every image
    were reprojected into the same all-sky map.

    The uncertainty layer of the NDCube is expected to be present and is handled.

    Parameters
    ----------
    data_cube : NDCube
        The data cube to be reprojected
    cdelt : float
        The CDELT value to use for the output projection

    Returns
    -------
    reprojected_cube : NDCube
        The cutout of the all-sky map containing the reprojected input data

    """
    # Combine the data and uncertainty layers along a new axis
    data = np.stack((data_cube.data, data_cube.uncertainty.array))

    wcs = data_cube.wcs.celestial

    # We build the output WCS here
    wcs_out = WCS(naxis=2)
    wcs_out.wcs.ctype = "RA---CAR", "DEC--CAR"
    wcs_out.wcs.cdelt = cdelt, cdelt
    # This won't be the exact center of the output image, but that's ok. Just need to be close.
    wcs_out.wcs.crpix = data.shape[-1] // 2, data.shape[-2] // 2
    crval = wcs.wcs.crval
    # Center our output frame near the input data. Since we choose an even multiple of CDELT, we'll be on the
    # same pixel grid as any other output from this function. The latitude value must be zero, otherwise the CAR
    # projection starts bending the lat/lon lines.
    wcs_out.wcs.crval = ((crval[0] // cdelt) * cdelt) % 360, 0

    # Now find the exact bounds of the input image in this output frame, so we can set the output array size
    xs = np.linspace(-1, data.shape[-1], 30)
    ys = np.linspace(-1, data.shape[-2], 30)
    edgex = np.concatenate((xs, np.full(len(ys), xs[-1]), xs, np.zeros(len(ys))))
    edgey = np.concatenate((np.zeros(len(xs)), ys, np.full(len(xs), ys[-1]), ys))

    xs, ys = utils.pixel_to_pixel(wcs, wcs_out, edgex, edgey)
    xmin, xmax = np.floor(xs.min()), np.ceil(xs.max())
    ymin, ymax = np.floor(ys.min()), np.ceil(ys.max())
    out_shape = int(ymax - ymin), int(xmax - xmin)

    # Shift crpix so that the bounding box we just found starts at pixel 0
    crpix = wcs_out.wcs.crpix
    wcs_out.wcs.crpix = crpix[0] - xmin, crpix[1] - ymin

    reprojected = reproject.reproject_adaptive((data, wcs.celestial), wcs_out, out_shape,
                                               return_footprint=False, roundtrip_coords=False,
                                               boundary_mode="strict", conserve_flux=True, center_jacobian=True)
    return NDCube(data=reprojected[0], uncertainty=StdDevUncertainty(reprojected[1]), wcs=wcs_out)


def shift_image_onto(source: NDCube,
                     target: NDCube,
                     fill_value: float = np.nan) -> NDCube:
    """
    Aligns one image to the frame of a second, if both are different cutouts of the same all-sky coordinate frame.

    These two input images should be outputs from `to_celestial_frame_cutout`. Co-aligning the images only requires
    cropping and padding the first image, since the two images are cutouts of the same pixel grid. This provides the
    same accuracy (to within ~floating-point error) as reprojecting the first image into the second's frame,
    while being incredibly faster.

    Input files must have two image dimensions, and can have arbitrarily-many extra leading dimensions.

    The uncertainty layer of the NDCube is expected to be present and is handled.

    Parameters
    ----------
    source : NDCube
        The image to the aligned
    target : NDCube
        The image onto which ``input`` is to be aligned
    fill_value : float
        The value to be used for empty pixels (pixels of the output frame that aren't spanned by the input image)

    Returns
    -------
    aligned_image : NDCube
        The data of the ``input`` in the WCS frame of ``target``.

    """
    if any(source.wcs.wcs.cdelt != target.wcs.wcs.cdelt):
        msg = "WCSes must have identical CDELTs"
        raise ValueError(msg)

    source_data = np.stack((source.data, source.uncertainty.array))

    # Record any amounts by which we'll have to pad the array
    padding = np.zeros((2, 2), dtype=int)

    corner = utils.pixel_to_pixel(target.wcs, source.wcs, target.data.shape[-1] - 1, target.data.shape[-2] - 1)
    # Sanity check---since the images are on the same pixel grids, the corner's coordinates should come out as integers
    np.testing.assert_allclose(np.rint(corner), corner)
    corner = np.rint(corner).astype(int)

    dcorner = corner[0] - (source_data.shape[-1] - 1), corner[1] - (source_data.shape[-2] - 1)
    if dcorner[0] < 0:
        source_data = source_data[..., :dcorner[0]]
    else:
        padding[1][1] = dcorner[0]
    if dcorner[1] < 0:
        source_data = source_data[..., :dcorner[1], :]
    else:
        padding[0][1] = dcorner[1]

    corner2 = utils.pixel_to_pixel(target.wcs, source.wcs, 0, 0)
    # Sanity check---since the images are on the same pixel grids, the corner's coordinates should come out as integers
    np.testing.assert_allclose(np.rint(corner2), corner2)
    corner2 = np.rint(corner2).astype(int)

    if corner2[0] > 0:
        source_data = source_data[..., :, corner2[0]:]
    else:
        padding[1][0] = -corner2[0]
    if corner2[1] > 0:
        source_data = source_data[..., corner2[1]:, :]
    else:
        padding[0][0] = -corner2[1]

    if corner2[0] > 0 and corner[0] < 0:
        # This can occur when there's no overlap between the two frames.
        padding[1] = [0, target.data.shape[-1]]

    if np.any(padding != 0):
        if source_data.ndim > 2:
            padding = np.concatenate(([(0, 0)] * (source_data.ndim - 2), padding), axis=0)
        source_data = np.pad(source_data, padding, constant_values=fill_value)
    return NDCube(data=source_data[0], wcs=target.wcs, uncertainty=StdDevUncertainty(source_data[1]))



from typing import Any

from ndcube import NDCube
from prefect import Flow, Task, flow, get_run_logger, task
from prefect.client.schemas.objects import TaskRun
from prefect.states import State
from prefect.variables import Variable

from punchbowl.data.punch_io import get_base_file_name, write_ndcube_to_fits


def completion_debugger(task: Task, task_run: TaskRun, state: State) -> None:
    """Run on task completion during debug mode."""
    if Variable.get("debug", False):
        cube = state.result()
        if isinstance(cube, NDCube):
            new_filename = f"{get_base_file_name(cube)}_{task.name}.fits"
            write_ndcube_to_fits(cube, new_filename, overwrite=True, write_hash=False)
        elif isinstance(cube, list):
            for i, c in enumerate(cube):
                new_filename = f"{get_base_file_name(c)}_{task.name}_{i}.fits"
                write_ndcube_to_fits(c, new_filename, overwrite=True, write_hash=False)
        else:
            logger = get_run_logger()
            logger.error(f"Cannot write debug output for {task} {task_run} in {state}.")

def failure_hook(task: Task, task_run: TaskRun, state: State) -> None:
    """Run if a punch_task fails."""

def punch_task(*args: Any, **kwargs: Any) -> Task:
    """Prefect task that does PUNCH special things."""
    return task(*args, **kwargs, on_completion=[completion_debugger], on_failure=[failure_hook])

def punch_flow(*args: Any, **kwargs: Any) -> Flow:
    """Prefect flow that does PUNCH special things."""
    return flow(*args, **kwargs, validate_parameters=False)
